# Copyright (c) 2018-2019 Robin Jarry
# SPDX-License-Identifier: MIT

import os
import shlex
from typing import List

import cffi


HERE = os.path.dirname(__file__)

BUILDER = cffi.FFI()
with open(os.path.join(HERE, "cdefs.h"), encoding="utf-8") as f:
    BUILDER.cdef(f.read())


def search_paths(env_var: str) -> List[str]:
    paths = []
    for p in os.environ.get(env_var, "").strip().split(":"):
        p = p.strip()
        if p:
            paths.append(p)
    return paths


HEADERS = search_paths("LIBYANG_HEADERS")
LIBRARIES = search_paths("LIBYANG_LIBRARIES")
EXTRA_CFLAGS = ["-Werror", "-std=c99"]
EXTRA_CFLAGS += shlex.split(os.environ.get("LIBYANG_EXTRA_CFLAGS", ""))
EXTRA_LDFLAGS = shlex.split(os.environ.get("LIBYANG_EXTRA_LDFLAGS", ""))

with open(os.path.join(HERE, "source.c"), encoding="utf-8") as f:
    BUILDER.set_source(
        "_libyang",
        f.read(),
        libraries=["yang"],
        extra_compile_args=EXTRA_CFLAGS,
        extra_link_args=EXTRA_LDFLAGS,
        include_dirs=HEADERS,
        library_dirs=LIBRARIES,
        py_limited_api=False,
    )

if __name__ == "__main__":
    BUILDER.compile()



# Copyright (c) 2020 6WIND S.A.
# Copyright (c) 2021 RACOM s.r.o.
# SPDX-License-Identifier: MIT

import logging
from typing import IO, Any, Dict, Iterator, Optional, Tuple, Union

from _libyang import ffi, lib
from .keyed_list import KeyedList
from .schema import (
    Module,
    SContainer,
    SLeaf,
    SLeafList,
    SList,
    SNode,
    SNotif,
    SRpc,
    Type,
)
from .util import DataType, IOType, LibyangError, c2str, ly_array_iter, str2c


LOG = logging.getLogger(__name__)


# -------------------------------------------------------------------------------------
def implicit_flags(
    no_config: bool = False,
    no_defaults: bool = False,
    no_state: bool = False,
    output: bool = False,
) -> int:
    flags = 0
    if no_config:
        flags |= lib.LYD_IMPLICIT_NO_CONFIG
    if no_state:
        flags |= lib.LYD_IMPLICIT_NO_STATE
    if no_defaults:
        flags |= lib.LYD_IMPLICIT_NO_DEFAULTS
    if output:
        flags |= lib.LYD_IMPLICIT_OUTPUT
    return flags


# -------------------------------------------------------------------------------------
def printer_flags(
    with_siblings: bool = False,
    pretty: bool = True,
    keep_empty_containers: bool = False,
    trim_default_values: bool = False,
    include_implicit_defaults: bool = False,
) -> int:
    flags = 0
    if with_siblings:
        flags |= lib.LYD_PRINT_SIBLINGS
    if not pretty:
        flags |= lib.LYD_PRINT_SHRINK
    if keep_empty_containers:
        flags |= lib.LYD_PRINT_EMPTY_CONT
    if trim_default_values:
        flags |= lib.LYD_PRINT_WD_TRIM
    if include_implicit_defaults:
        flags |= lib.LYD_PRINT_WD_ALL
    return flags


# -------------------------------------------------------------------------------------
def data_format(fmt_string: str) -> int:
    if fmt_string == "json":
        return lib.LYD_JSON
    if fmt_string == "xml":
        return lib.LYD_XML
    if fmt_string == "lyb":
        return lib.LYD_LYB
    raise ValueError("unknown data format: %r" % fmt_string)


# -------------------------------------------------------------------------------------
def newval_flags(
    rpc_output: bool = False,
    store_only: bool = False,
    bin_value: bool = False,
    canon_value: bool = False,
    meta_clear_default: bool = False,
    update: bool = False,
    opaq: bool = False,
) -> int:
    """
    Translate from booleans to newvaloptions flags.
    """
    flags = 0
    if rpc_output:
        flags |= lib.LYD_NEW_VAL_OUTPUT
    if store_only:
        flags |= lib.LYD_NEW_VAL_STORE_ONLY
    if bin_value:
        flags |= lib.LYD_NEW_VAL_BIN
    if canon_value:
        flags |= lib.LYD_NEW_VAL_CANON
    if meta_clear_default:
        flags |= lib.LYD_NEW_META_CLEAR_DFLT
    if update:
        flags |= lib.LYD_NEW_PATH_UPDATE
    if opaq:
        flags |= lib.LYD_NEW_PATH_OPAQ
    return flags


# -------------------------------------------------------------------------------------
def parser_flags(
    no_state: bool = False,
    parse_only: bool = False,
    opaq: bool = False,
    ordered: bool = False,
    strict: bool = False,
    store_only: bool = False,
    json_null: bool = False,
    json_string_datatypes: bool = False,
) -> int:
    flags = 0
    if no_state:
        flags |= lib.LYD_PARSE_NO_STATE
    if parse_only:
        flags |= lib.LYD_PARSE_ONLY
    if opaq:
        flags |= lib.LYD_PARSE_OPAQ
    if ordered:
        flags |= lib.LYD_PARSE_ORDERED
    if strict:
        flags |= lib.LYD_PARSE_STRICT
    if store_only:
        flags |= lib.LYD_PARSE_STORE_ONLY
    if json_null:
        flags |= lib.LYD_PARSE_JSON_NULL
    if json_string_datatypes:
        flags |= lib.LYD_PARSE_JSON_STRING_DATATYPES
    return flags


# -------------------------------------------------------------------------------------
def merge_flags(
    defaults: bool = False,
    destruct: bool = False,
    with_flags: bool = False,
) -> int:
    flags = 0
    if defaults:
        flags |= lib.LYD_MERGE_DEFAULTS
    if destruct:
        flags |= lib.LYD_MERGE_DESTRUCT
    if with_flags:
        flags |= lib.LYD_MERGE_WITH_FLAGS
    return flags


# -------------------------------------------------------------------------------------
def dup_flags(
    no_meta: bool = False,
    recursive: bool = False,
    with_flags: bool = False,
    with_parents: bool = False,
) -> int:
    flags = 0
    if no_meta:
        flags |= lib.LYD_DUP_NO_META
    if recursive:
        flags |= lib.LYD_DUP_RECURSIVE
    if with_flags:
        flags |= lib.LYD_DUP_WITH_FLAGS
    if with_parents:
        flags |= lib.LYD_DUP_WITH_PARENTS
    return flags


# -------------------------------------------------------------------------------------
def data_type(dtype: DataType) -> int:
    if not isinstance(dtype, DataType):
        dtype = DataType(dtype)
    return dtype.value


# -------------------------------------------------------------------------------------
def validation_flags(
    no_state: bool = False,
    validate_present: bool = False,
    validate_multi_error: bool = False,
) -> int:
    flags = 0
    if no_state:
        flags |= lib.LYD_VALIDATE_NO_STATE
    if validate_present:
        flags |= lib.LYD_VALIDATE_PRESENT
    if validate_multi_error:
        flags |= lib.LYD_VALIDATE_MULTI_ERROR
    return flags


def diff_flags(with_defaults: bool = False) -> int:
    flags = 0
    if with_defaults:
        flags |= lib.LYD_DIFF_DEFAULTS
    return flags


# -------------------------------------------------------------------------------------
class DNodeAttrs:
    __slots__ = ("context", "parent", "cdata", "__dict__")

    def __init__(self, context: "libyang.Context", parent: "libyang.DNode"):
        self.context = context
        self.parent = parent
        self.cdata = []  # C type: "struct lyd_attr *"

    def get(self, name: str) -> Optional[str]:
        for attr_name, attr_value in self:
            if attr_name == name:
                return attr_value
        return None

    def set(self, name: str, value: str):
        attrs = ffi.new("struct lyd_attr **")
        ret = lib.lyd_new_attr(
            self.parent.cdata,
            ffi.NULL,
            str2c(name),
            str2c(value),
            attrs,
        )
        if ret != lib.LY_SUCCESS:
            raise self.context.error("cannot create attr")
        self.cdata.append(attrs[0])

    def remove(self, name: str):
        for attr in self.cdata:
            if self._get_attr_name(attr) == name:
                lib.lyd_free_attr_single(self.context.cdata, attr)
                self.cdata.remove(attr)
                break

    def __contains__(self, name: str) -> bool:
        for attr_name, _ in self:
            if attr_name == name:
                return True
        return False

    def __iter__(self) -> Iterator[Tuple[str, str]]:
        for attr in self.cdata:
            name = self._get_attr_name(attr)
            yield (name, c2str(attr.value))

    def __len__(self) -> int:
        return len(self.cdata)

    @staticmethod
    def _get_attr_name(cdata) -> str:
        if cdata.name.prefix != ffi.NULL:
            return f"{c2str(cdata.name.prefix)}:{c2str(cdata.name.name)}"
        return c2str(cdata.name.name)


# -------------------------------------------------------------------------------------
class DNode:
    """
    Data tree node.
    """

    __slots__ = ("context", "cdata", "attributes", "free_func", "__dict__")

    def __init__(self, context: "libyang.Context", cdata):
        """
        :arg context:
            The libyang.Context python object.
        :arg cdata:
            The pointer to the C structure allocated by libyang.so.
        """
        self.context = context
        self.cdata = cdata  # C type: "struct lyd_node *"
        self.attributes = None
        self.free_func = None  # type: Callable[DNode]

    def meta(self) -> Dict[str, str]:
        ret = {}
        item = self.cdata.meta
        while item != ffi.NULL:
            ret[c2str(item.name)] = c2str(
                lib.lyd_value_get_canonical(
                    self.context.cdata, ffi.addressof(item.value)
                )
            )
            item = item.next
        return ret

    def get_meta(self, name: str) -> Optional[str]:
        item = self.cdata.meta
        while item != ffi.NULL:
            if c2str(item.name) == name:
                return c2str(
                    lib.lyd_value_get_canonical(
                        self.context.cdata, ffi.addressof(item.value)
                    )
                )
            item = item.next
        return None

    def meta_free(self, name: str):
        item = self.cdata.meta
        while item != ffi.NULL:
            if c2str(item.name) == name:
                lib.lyd_free_meta_single(item)
                break
            item = item.next

    def new_meta(
        self, name: str, value: str, clear_dflt: bool = False, store_only: bool = False
    ):
        flags = newval_flags(meta_clear_default=clear_dflt, store_only=store_only)
        ret = lib.lyd_new_meta(
            ffi.NULL,
            self.cdata,
            ffi.NULL,
            str2c(name),
            str2c(value),
            flags,
            ffi.NULL,
        )
        if ret != lib.LY_SUCCESS:
            raise self.context.error("cannot create meta")

    def attrs(self) -> DNodeAttrs:
        if not self.attributes:
            self.attributes = DNodeAttrs(self.context, self)
        return self.attributes

    def add_defaults(
        self,
        no_config: bool = False,
        no_defaults: bool = False,
        no_state: bool = False,
        output: bool = False,
        only_node: bool = False,
        only_module: Optional[Module] = None,
    ):
        flags = implicit_flags(
            no_config=no_config,
            no_defaults=no_defaults,
            no_state=no_state,
            output=output,
        )
        if only_node:
            node_p = ffi.cast("struct lyd_node *", self.cdata)
            ret = lib.lyd_new_implicit_tree(node_p, flags, ffi.NULL)
        else:
            node_p = ffi.new("struct lyd_node **")
            node_p[0] = self.cdata
            if only_module is not None:
                ret = lib.lyd_new_implicit_module(
                    node_p, only_module.cdata, flags, ffi.NULL
                )
            else:
                ret = lib.lyd_new_implicit_all(
                    node_p, self.context.cdata, flags, ffi.NULL
                )

        if ret != lib.LY_SUCCESS:
            raise self.context.error("cannot get module")

    def flags(self):
        ret = {"default": False, "when_true": False, "new": False}
        if self.cdata.flags & lib.LYD_DEFAULT:
            ret["default"] = True
        if self.cdata.flags & lib.LYD_WHEN_TRUE:
            ret["when_true"] = True
        if self.cdata.flags & lib.LYD_NEW:
            ret["new"] = True
        return ret

    def is_default(self) -> bool:
        return lib.lyd_is_default(self.cdata)

    def set_when(self, value: bool):
        if value:
            self.cdata.flags |= lib.LYD_WHEN_TRUE
        else:
            self.cdata.flags &= ~lib.LYD_WHEN_TRUE

    def new_path(
        self,
        path: str,
        value: str,
        opt_update: bool = False,
        opt_output: bool = False,
        opt_opaq: bool = False,
        opt_bin_value: bool = False,
        opt_canon_value: bool = False,
        opt_store_only: bool = False,
    ):
        flags = newval_flags(
            update=opt_update,
            rpc_output=opt_output,
            opaq=opt_opaq,
            bin_value=opt_bin_value,
            canon_value=opt_canon_value,
            store_only=opt_store_only,
        )
        ret = lib.lyd_new_path(
            self.cdata, ffi.NULL, str2c(path), str2c(value), flags, ffi.NULL
        )
        if ret != lib.LY_SUCCESS:
            raise self.context.error("cannot get module")

    def insert_child(self, node):
        ret = lib.lyd_insert_child(self.cdata, node.cdata)
        if ret != lib.LY_SUCCESS:
            raise self.context.error("cannot insert node")

    def insert_sibling(self, node):
        ret = lib.lyd_insert_sibling(self.cdata, node.cdata, ffi.NULL)
        if ret != lib.LY_SUCCESS:
            raise self.context.error("cannot insert sibling")

    def insert_after(self, node):
        ret = lib.lyd_insert_after(self.cdata, node.cdata)
        if ret != lib.LY_SUCCESS:
            raise self.context.error("cannot insert sibling after")

    def insert_before(self, node):
        ret = lib.lyd_insert_before(self.cdata, node.cdata)
        if ret != lib.LY_SUCCESS:
            raise self.context.error("cannot insert sibling before")

    def name(self) -> str:
        return c2str(self.cdata.schema.name)

    def module(self) -> Module:
        if not self.cdata.schema:
            raise self.context.error("cannot get module")
        return Module(self.context, self.cdata.schema.module)

    def schema(self) -> SNode:
        return SNode.new(self.context, self.cdata.schema)

    def parent(self) -> Optional["DNode"]:
        if not self.cdata.parent:
            return None
        return self.new(self.context, self.cdata.parent)

    def next(self) -> Optional["DNode"]:
        if not self.cdata.next:
            return None
        return self.new(self.context, self.cdata.next)

    def prev(self) -> Optional["DNode"]:
        if not self.cdata.prev:
            return None
        return self.new(self.context, self.cdata.prev)

    def root(self) -> "DNode":
        node = self
        while node.parent() is not None:
            node = node.parent()
        return node

    def first_sibling(self) -> "DNode":
        n = lib.lyd_first_sibling(self.cdata)
        if n == self.cdata:
            return self
        return self.new(self.context, n)

    def siblings(self, include_self: bool = True) -> Iterator["DNode"]:
        n = lib.lyd_first_sibling(self.cdata)
        while n:
            if n == self.cdata:
                if include_self:
                    yield self
            else:
                yield self.new(self.context, n)
            n = n.next

    def find_path(self, path: str, output: bool = False):
        node = ffi.new("struct lyd_node **")
        ret = lib.lyd_find_path(self.cdata, str2c(path), output, node)
        if ret == lib.LY_SUCCESS:
            return DNode.new(self.context, node[0])
        return None

    def find_one(self, xpath: str) -> Optional["DNode"]:
        try:
            return next(self.find_all(xpath))
        except StopIteration:
            return None

    def find_all(self, xpath: str) -> Iterator["DNode"]:
        node_set = ffi.new("struct ly_set **")
        ret = lib.lyd_find_xpath(self.cdata, str2c(xpath), node_set)
        if ret != lib.LY_SUCCESS:
            raise self.context.error("cannot find path: %s", xpath)

        node_set = node_set[0]
        try:
            for i in range(node_set.count):
                n = DNode.new(self.context, node_set.dnodes[i])
                yield n
        finally:
            lib.ly_set_free(node_set, ffi.NULL)

    def eval_xpath(self, xpath: str):
        lbool = ffi.new("ly_bool *")
        ret = lib.lyd_eval_xpath(self.cdata, str2c(xpath), lbool)
        if ret != lib.LY_SUCCESS:
            raise self.context.error("cannot eva xpath: %s", xpath)
        if lbool[0]:
            return True
        return False

    def path(self) -> str:
        return self._get_path(self.cdata)

    def validate(
        self,
        no_state: bool = False,
        validate_present: bool = False,
        rpc: bool = False,
        rpcreply: bool = False,
        notification: bool = False,
        dep_tree: Optional["DNode"] = None,
    ) -> None:
        dtype = None
        if rpc:
            dtype = DataType.RPC_YANG
        elif rpcreply:
            dtype = DataType.REPLY_YANG
        elif notification:
            dtype = DataType.NOTIF_YANG

        if dtype is None:
            self.validate_all(no_state, validate_present)
        else:
            self.validate_op(dtype, dep_tree)

    def validate_all(
        self,
        no_state: bool = False,
        validate_present: bool = False,
    ) -> None:
        if self.cdata.parent:
            raise self.context.error("validation is only supported on top-level nodes")
        flags = validation_flags(
            no_state=no_state,
            validate_present=validate_present,
        )
        node_p = ffi.new("struct lyd_node **")
        node_p[0] = self.cdata
        ret = lib.lyd_validate_all(node_p, self.context.cdata, flags, ffi.NULL)
        if ret != lib.LY_SUCCESS:
            raise self.context.error("validation failed")

    def validate_op(
        self,
        dtype: DataType,
        dep_tree: Optional["DNode"] = None,
    ) -> None:
        dtype = data_type(dtype)
        ret = lib.lyd_validate_op(
            self.cdata,
            ffi.NULL if dep_tree is None else dep_tree.cdata,
            dtype,
            ffi.NULL,
        )
        if ret != lib.LY_SUCCESS:
            raise self.context.error("validation failed")

    def diff(
        self,
        other: "DNode",
        no_siblings: bool = False,
        with_defaults: bool = False,
    ) -> "DNode":
        flags = diff_flags(with_defaults=with_defaults)
        node_p = ffi.new("struct lyd_node **")
        if no_siblings:
            ret = lib.lyd_diff_tree(self.cdata, other.cdata, flags, node_p)
        else:
            ret = lib.lyd_diff_siblings(self.cdata, other.cdata, flags, node_p)

        if ret != lib.LY_SUCCESS:
            raise self.context.error("diff error")

        if node_p[0] == ffi.NULL:
            return None

        return self.new(self.context, node_p[0])

    def diff_apply(self, diff_node: "DNode") -> None:
        node_p = ffi.new("struct lyd_node **")
        node_p[0] = self.cdata

        ret = lib.lyd_diff_apply_all(node_p, diff_node.cdata)
        if ret != lib.LY_SUCCESS:
            raise self.context.error("apply diff error")

    def duplicate(
        self,
        with_siblings: bool = False,
        no_meta: bool = False,
        recursive: bool = False,
        with_flags: bool = False,
        with_parents: bool = False,
        parent: Optional["DNode"] = None,
    ) -> "DNode":
        flags = dup_flags(
            no_meta=no_meta,
            recursive=recursive,
            with_flags=with_flags,
            with_parents=with_parents,
        )

        if parent is not None:
            parent = parent.cdata
        else:
            parent = ffi.NULL

        node = ffi.new("struct lyd_node **")
        if with_siblings:
            lib.lyd_dup_siblings(self.cdata, parent, flags, node)
        else:
            lib.lyd_dup_single(self.cdata, parent, flags, node)

        return DNode.new(self.context, node[0])

    def merge_module(
        self,
        source: "DNode",
        defaults: bool = False,
        destruct: bool = False,
        with_flags: bool = False,
    ) -> None:
        flags = merge_flags(defaults=defaults, destruct=destruct, with_flags=with_flags)
        node_p = ffi.new("struct lyd_node **")
        node_p[0] = self.cdata
        ret = lib.lyd_merge_module(
            node_p, source.cdata, ffi.NULL, ffi.NULL, ffi.NULL, flags
        )
        if ret != lib.LY_SUCCESS:
            raise self.context.error("merge failed")

    def merge(
        self,
        source: "DNode",
        with_siblings: bool = False,
        defaults: bool = False,
        destruct: bool = False,
        with_flags: bool = False,
    ) -> None:
        flags = merge_flags(defaults=defaults, destruct=destruct, with_flags=with_flags)
        node_p = ffi.new("struct lyd_node **")
        node_p[0] = self.cdata
        if with_siblings:
            ret = lib.lyd_merge_siblings(node_p, source.cdata, flags)
        else:
            ret = lib.lyd_merge_tree(node_p, source.cdata, flags)

        if ret != lib.LY_SUCCESS:
            raise self.context.error("merge failed")

        self.cdata = node_p[0]

    def iter_tree(self) -> Iterator["DNode"]:
        n = next_n = self.cdata
        while n != ffi.NULL:
            yield self.new(self.context, n)

            next_n = lib.lyd_child(n)
            if next_n == ffi.NULL:
                if n == self.cdata:
                    break
                next_n = n.next
            while next_n == ffi.NULL:
                n = n.parent
                if n.parent == self.cdata.parent:
                    break
                next_n = n.next
            n = next_n

    def print(
        self,
        fmt: str,
        out_type: IOType,
        out_target: Union[IO, str, None] = None,
        with_siblings: bool = False,
        pretty: bool = True,
        keep_empty_containers: bool = False,
        trim_default_values: bool = False,
        include_implicit_defaults: bool = False,
    ) -> Union[str, bytes]:
        flags = printer_flags(
            pretty=pretty,
            keep_empty_containers=keep_empty_containers,
            trim_default_values=trim_default_values,
            include_implicit_defaults=include_implicit_defaults,
        )
        fmt = data_format(fmt)
        out_data = ffi.new("struct ly_out **")

        if out_type == IOType.FD:
            raise NotImplementedError

        if out_type == IOType.FILE:
            raise NotImplementedError

        if out_type == IOType.FILEPATH:
            raise NotImplementedError

        if out_type != IOType.MEMORY:
            raise ValueError("no input specified")

        buf = ffi.new("char **")
        ret = lib.ly_out_new_memory(buf, 0, out_data)
        if ret != lib.LY_SUCCESS:
            raise self.context.error("failed to initialize output target")

        if with_siblings:
            ret = lib.lyd_print_all(out_data[0], self.cdata, fmt, flags)
        else:
            ret = lib.lyd_print_tree(out_data[0], self.cdata, fmt, flags)

        lib.ly_out_free(out_data[0], ffi.NULL, 0)
        if ret != lib.LY_SUCCESS:
            raise self.context.error("failed to write data")

        return c2str(buf[0], decode=True)

    def print_mem(
        self,
        fmt: str,
        with_siblings: bool = False,
        pretty: bool = True,
        keep_empty_containers: bool = False,
        trim_default_values: bool = False,
        include_implicit_defaults: bool = False,
    ) -> Union[str, bytes]:
        flags = printer_flags(
            with_siblings=with_siblings,
            pretty=pretty,
            keep_empty_containers=keep_empty_containers,
            trim_default_values=trim_default_values,
            include_implicit_defaults=include_implicit_defaults,
        )
        buf = ffi.new("char **")
        fmt = data_format(fmt)
        ret = lib.lyd_print_mem(buf, self.cdata, fmt, flags)
        if ret != lib.LY_SUCCESS:
            raise self.context.error("cannot print node")
        try:
            if fmt == lib.LYD_LYB:
                # binary format, do not convert to unicode
                return c2str(buf[0], decode=False)
            return c2str(buf[0], decode=True)
        finally:
            lib.free(buf[0])

    def print_file(
        self,
        fileobj: IO,
        fmt: str,
        with_siblings: bool = False,
        pretty: bool = True,
        keep_empty_containers: bool = False,
        trim_default_values: bool = False,
        include_implicit_defaults: bool = False,
    ) -> None:
        flags = printer_flags(
            with_siblings=with_siblings,
            pretty=pretty,
            keep_empty_containers=keep_empty_containers,
            trim_default_values=trim_default_values,
            include_implicit_defaults=include_implicit_defaults,
        )
        fmt = data_format(fmt)
        out = ffi.new("struct ly_out **")
        ret = lib.ly_out_new_fd(fileobj.fileno(), out)
        if ret != lib.LY_SUCCESS:
            raise self.context.error("cannot allocate output data")
        if with_siblings:
            ret = lib.lyd_print_all(out[0], self.cdata, fmt, flags)
        else:
            ret = lib.lyd_print_tree(out[0], self.cdata, fmt, flags)
        if ret != lib.LY_SUCCESS:
            raise self.context.error("cannot print node")

    def should_print(
        self,
        include_implicit_defaults: bool = False,
        trim_default_values: bool = False,
        keep_empty_containers: bool = False,
    ) -> bool:
        """
        Check if a node should be "printed".

        :arg bool include_implicit_defaults:
            Include implicit default nodes.
        :arg bool trim_default_values:
            Exclude nodes with the value equal to their default value.
        :arg bool keep_empty_containers:
            Preserve empty non-presence containers.
        """
        flags = printer_flags(
            include_implicit_defaults=include_implicit_defaults,
            trim_default_values=trim_default_values,
            keep_empty_containers=keep_empty_containers,
        )
        return bool(lib.lyd_node_should_print(self.cdata, flags))

    def print_dict(
        self,
        strip_prefixes: bool = True,
        absolute: bool = True,
        with_siblings: bool = False,
        include_implicit_defaults: bool = False,
        trim_default_values: bool = False,
        keep_empty_containers: bool = False,
    ) -> Dict[str, Any]:
        """
        Convert a DNode object to a python dictionary.

        The format is inspired by the YANG JSON format described in :rfc:`7951` but has
        some differences:

        * ``int64`` and ``uint64`` values are represented by python ``int`` values
          instead of string values.
        * ``decimal64`` values are represented by python ``float`` values instead of
          string values.
        * ``empty`` values are represented by python ``None`` values instead of
          ``[None]`` list instances. To check if an ``empty`` leaf is set in a
          container, you should use the idiomatic ``if "foo" in container:`` construct.

        :arg bool strip_prefixes:
            If True (the default), module prefixes are stripped from dictionary keys. If
            False, dictionary keys are in the form ``<module>:<name>``.
        :arg bool absolute:
            If True (the default), always return a dictionary containing the complete
            tree starting from the root.
        :arg bool with_siblings:
            If True, include the node's siblings.
        :arg bool include_implicit_defaults:
            Include implicit default nodes.
        :arg bool trim_default_values:
            Exclude nodes with the value equal to their default value.
        :arg bool keep_empty_containers:
            Preserve empty non-presence containers.
        """
        flags = printer_flags(
            include_implicit_defaults=include_implicit_defaults,
            trim_default_values=trim_default_values,
            keep_empty_containers=keep_empty_containers,
        )

        name_cache = {}

        def _node_name(node):
            name = name_cache.get(node.schema)
            if name is None:
                if strip_prefixes:
                    name = c2str(node.schema.name)
                else:
                    mod = node.schema.module
                    name = "%s:%s" % (c2str(mod.name), c2str(node.schema.name))
                name_cache[node.schema] = name
            return name

        list_keys_cache = {}

        def _init_yang_list(snode):
            if snode.flags & lib.LYS_ORDBY_USER:
                return []  # ordered list, return an empty builtin list

            # unordered lists
            if snode.nodetype == SNode.LEAFLIST:
                return KeyedList(key_name=None)

            if snode not in list_keys_cache:
                keys = []
                child = lib.lysc_node_child(snode)
                while child:
                    if child.flags & lib.LYS_KEY:
                        if strip_prefixes:
                            keys.append(c2str(child.name))
                        else:
                            keys.append(
                                "%s:%s" % (c2str(child.module.name), c2str(child.name))
                            )
                    child = child.next
                if len(keys) == 1:
                    list_keys_cache[snode] = keys[0]
                else:
                    list_keys_cache[snode] = tuple(keys)

            return KeyedList(key_name=list_keys_cache[snode])

        def _to_dict(node, parent_dic):
            if not lib.lyd_node_should_print(node, flags):
                return
            name = _node_name(node)
            if node.schema.nodetype == SNode.LIST:
                list_element = {}

                child = lib.lyd_child(node)
                while child:
                    _to_dict(child, list_element)
                    child = child.next
                if name not in parent_dic:
                    parent_dic[name] = _init_yang_list(node.schema)
                parent_dic[name].append(list_element)
            elif node.schema.nodetype & (
                SNode.CONTAINER | SNode.RPC | SNode.ACTION | SNode.NOTIF
            ):
                container = {}
                child = lib.lyd_child(node)
                while child:
                    _to_dict(child, container)
                    child = child.next
                parent_dic[name] = container
            elif node.schema.nodetype == SNode.LEAFLIST:
                if name not in parent_dic:
                    parent_dic[name] = _init_yang_list(node.schema)
                parent_dic[name].append(DLeaf.cdata_leaf_value(node, self.context))
            elif node.schema.nodetype == SNode.LEAF:
                parent_dic[name] = DLeaf.cdata_leaf_value(node, self.context)

        dic = {}
        dnode = self
        if absolute:
            dnode = dnode.root()
        if with_siblings:
            for sib in dnode.siblings():
                _to_dict(sib.cdata, dic)
        else:
            _to_dict(dnode.cdata, dic)
        return dic

    def merge_data_dict(
        self,
        dic: Dict[str, Any],
        no_state: bool = False,
        validate_present: bool = False,
        validate: bool = True,
        strict: bool = False,
        rpc: bool = False,
        rpcreply: bool = False,
    ) -> Optional["DNode"]:
        """
        Merge a python dictionary into this node. The returned value is the first
        created node.

        :arg dic:
            The python dictionary to convert.
        :arg no_state:
            Consider state data not allowed and raise an error during validation if they are found.
        :arg validate_present:
            Validate result of the operation against schema.
        :arg validate:
            Run validation on result of the operation.
        :arg strict:
            Instead of ignoring data without schema definition, raise an error.
        :arg rpc:
            Data represents RPC or action input parameters.
        :arg rpcreply:
            Data represents RPC or action output parameters.
        """
        return dict_to_dnode(
            dic,
            self.module(),
            parent=self,
            no_state=no_state,
            validate_present=validate_present,
            validate=validate,
            strict=strict,
            rpc=rpc,
            rpcreply=rpcreply,
        )

    def unlink(self, with_siblings: bool = False) -> None:
        if with_siblings:
            lib.lyd_unlink_siblings(self.cdata)
        else:
            lib.lyd_unlink_tree(self.cdata)

    def free_internal(self, with_siblings: bool = True) -> None:
        if with_siblings:
            lib.lyd_free_all(self.cdata)
        else:
            lib.lyd_free_tree(self.cdata)

    def free(self, with_siblings: bool = True) -> None:
        try:
            if self.free_func:
                self.free_func(self)  # pylint: disable=not-callable
            else:
                self.free_internal(with_siblings)
        finally:
            self.cdata = ffi.NULL

    def leafref_link_node_tree(self) -> None:
        """
        Traverse through data tree including root node siblings and adds
        leafrefs links to the given nodes.

        Requires leafref_linking to be set on the libyang context.
        """
        lib.lyd_leafref_link_node_tree(self.cdata)

    def leafref_nodes(self) -> Iterator["DNode"]:
        """
        Gets the nodes that are referring to this node.

        Requires leafref_linking to be set on the libyang context.
        """
        term_node = ffi.cast("struct lyd_node_term *", self.cdata)
        out = ffi.new("const struct lyd_leafref_links_rec **")
        if lib.lyd_leafref_get_links(term_node, out) != lib.LY_SUCCESS:
            return
        for n in ly_array_iter(out[0].leafref_nodes):
            yield DNode.new(self.context, n)

    def target_nodes(self) -> Iterator["DNode"]:
        """
        Gets the target nodes that are referred by this node.

        Requires leafref_linking to be set on the libyang context.
        """
        term_node = ffi.cast("struct lyd_node_term *", self.cdata)
        out = ffi.new("const struct lyd_leafref_links_rec **")
        if lib.lyd_leafref_get_links(term_node, out) != lib.LY_SUCCESS:
            return
        for n in ly_array_iter(out[0].target_nodes):
            yield DNode.new(self.context, n)

    def __repr__(self):
        cls = self.__class__
        return "<%s.%s: %s>" % (cls.__module__, cls.__name__, str(self))

    def __str__(self):
        return self.name()

    NODETYPE_CLASS = {}

    @classmethod
    def register(cls, *nodetypes):
        def _decorator(nodeclass):
            for t in nodetypes:
                cls.NODETYPE_CLASS[t] = nodeclass
            return nodeclass

        return _decorator

    @classmethod
    def new(cls, context: "libyang.Context", cdata) -> "DNode":
        cdata = ffi.cast("struct lyd_node *", cdata)
        if not cdata.schema:
            schemas = list(context.find_path(cls._get_path(cdata)))
            if len(schemas) != 1:
                raise LibyangError("Unable to determine schema")
            nodecls = cls.NODETYPE_CLASS.get(schemas[0].nodetype(), None)
        else:
            nodecls = cls.NODETYPE_CLASS.get(cdata.schema.nodetype, None)
        if nodecls is None:
            raise TypeError("node type %s not implemented" % cdata.schema.nodetype)
        return nodecls(context, cdata)

    @staticmethod
    def _get_path(cdata) -> str:
        path = lib.lyd_path(cdata, lib.LYD_PATH_STD, ffi.NULL, 0)
        try:
            return c2str(path)
        finally:
            lib.free(path)


# -------------------------------------------------------------------------------------
@DNode.register(SNode.CONTAINER)
class DContainer(DNode):
    def create_path(
        self,
        path: str,
        value: Any = None,
        rpc_output: bool = False,
        store_only: bool = False,
    ) -> Optional[DNode]:
        return self.context.create_data_path(
            path, parent=self, value=value, rpc_output=rpc_output, store_only=store_only
        )

    def children(self, no_keys=False) -> Iterator[DNode]:
        if no_keys:
            child = lib.lyd_child_no_keys(self.cdata)
        else:
            child = lib.lyd_child(self.cdata)

        while child:
            if child.schema != ffi.NULL:
                yield DNode.new(self.context, child)
            child = child.next

    def __iter__(self):
        return self.children()


# -------------------------------------------------------------------------------------
@DNode.register(SNode.RPC)
@DNode.register(SNode.ACTION)
class DRpc(DContainer):
    pass


# -------------------------------------------------------------------------------------
@DNode.register(SNode.LIST)
class DList(DContainer):
    pass


# -------------------------------------------------------------------------------------
@DNode.register(SNode.LEAF)
class DLeaf(DNode):
    def value(self) -> Any:
        return DLeaf.cdata_leaf_value(self.cdata, self.context)

    @staticmethod
    def cdata_leaf_value(cdata, context: "libyang.Context" = None) -> Any:
        val = lib.lyd_get_value(cdata)
        if val == ffi.NULL:
            return None

        val = c2str(val)
        if cdata.schema == ffi.NULL:
            # opaq node
            return val

        node_term = ffi.cast("struct lyd_node_term *", cdata)

        # inspired from libyang lyd_value_validate
        val_type = Type(context, node_term.value.realtype, None).base()
        if val_type == Type.UNION:
            val_type = Type(
                context, node_term.value.subvalue.value.realtype, None
            ).base()

        if val_type in Type.STR_TYPES:
            return val
        if val_type in Type.NUM_TYPES:
            return int(val)
        if val_type == Type.BOOL:
            return val == "true"
        if val_type == Type.DEC64:
            return float(val)
        if val_type == Type.EMPTY:
            return None
        return val


# -------------------------------------------------------------------------------------
@DNode.register(SNode.LEAFLIST)
class DLeafList(DLeaf):
    pass


# -------------------------------------------------------------------------------------
@DNode.register(SNode.NOTIF)
class DNotif(DContainer):
    pass


# -------------------------------------------------------------------------------------
@DNode.register(SNode.ANYXML)
class DAnyxml(DNode):
    def value(self, fmt: str = "xml"):
        anystr = ffi.new("char **", ffi.NULL)
        ret = lib.lyd_any_value_str(self.cdata, anystr)
        if ret != lib.LY_SUCCESS:
            raise self.context.error("cannot get data")
        return c2str(anystr[0])


# -------------------------------------------------------------------------------------
@DNode.register(SNode.ANYDATA)
class DAnydata(DNode):
    def value(self, fmt: str = "xml"):
        anystr = ffi.new("char **", ffi.NULL)
        ret = lib.lyd_any_value_str(self.cdata, anystr)
        if ret != lib.LY_SUCCESS:
            raise self.context.error("cannot get data")
        return c2str(anystr[0])


# -------------------------------------------------------------------------------------
def dict_to_dnode(
    dic: Dict[str, Any],
    module: Module,
    parent: Optional[DNode] = None,
    no_state: bool = False,
    validate_present: bool = False,
    validate: bool = True,
    strict: bool = False,
    rpc: bool = False,
    rpcreply: bool = False,
    notification: bool = False,
    store_only: bool = False,
) -> Optional[DNode]:
    """
    Convert a python dictionary to a DNode object given a YANG module object. The return
    value is the first created node. If parent is not set, a top-level node is returned.

    :arg dic:
        The python dictionary to convert.
    :arg module:
        The libyang Module object associated with the dictionary.
    :arg parent:
        Optional parent to update. If not specified a new top-level DNode will be
        created.
    :arg no_state:
        Consider state data not allowed and raise an error during validation if they are found.
    :arg validate_present:
        Validate result of the operation against schema.
    :arg validate:
        Run validation on result of the operation.
    :arg strict:
        Instead of ignoring data without schema definition, raise an error.
    :arg rpc:
        Data represents RPC or action input parameters.
    :arg rpcreply:
        Data represents RPC or action output parameters.
    :arg notification:
        Data represents notification parameters.
    :arg store_only:
        Data are being stored regardless of type validation (length, range, pattern, etc.)
    """
    if not dic:
        return None

    if not isinstance(dic, dict):
        raise TypeError("dic argument must be a python dict")
    if not isinstance(module, Module):
        raise TypeError("module argument must be a Module object")
    if parent is not None and not isinstance(parent, DNode):
        raise TypeError("parent argument must be a DNode object or None")

    created = []

    def _create_leaf(_parent, module, name, value, in_rpc_output=False):
        if value is not None:
            if isinstance(value, bool):
                value = str(value).lower()
            elif not isinstance(value, str):
                value = str(value)

        n = ffi.new("struct lyd_node **")
        flags = newval_flags(rpc_output=in_rpc_output, store_only=store_only)
        ret = lib.lyd_new_term(
            _parent,
            module.cdata,
            str2c(name),
            str2c(value),
            flags,
            n,
        )

        if ret != lib.LY_SUCCESS:
            if _parent:
                parent_path = repr(DNode.new(module.context, _parent).path())
            else:
                parent_path = "module %r" % module.name()
            raise module.context.error(
                "failed to create leaf %r as a child of %s", name, parent_path
            )
        created.append(n[0])

    def _create_container(_parent, module, name, in_rpc_output=False):
        n = ffi.new("struct lyd_node **")
        ret = lib.lyd_new_inner(_parent, module.cdata, str2c(name), in_rpc_output, n)
        if ret != lib.LY_SUCCESS:
            if _parent:
                parent_path = repr(DNode.new(module.context, _parent).path())
            else:
                parent_path = "module %r" % module.name()
            raise module.context.error(
                "failed to create container/list/rpc %r as a child of %s",
                name,
                parent_path,
            )
        created.append(n[0])
        return n[0]

    def _create_list(_parent, module, name, key_values, in_rpc_output=False):
        n = ffi.new("struct lyd_node **")
        flags = newval_flags(rpc_output=in_rpc_output, store_only=store_only)
        ret = lib.lyd_new_list(
            _parent,
            module.cdata,
            str2c(name),
            flags,
            n,
            *[str2c(str(i)) for i in key_values],
        )
        if ret != lib.LY_SUCCESS:
            if _parent:
                parent_path = repr(DNode.new(module.context, _parent).path())
            else:
                parent_path = "module %r" % module.name()
            raise module.context.error(
                "failed to create container/list/rpc %r as a child of %s",
                name,
                parent_path,
            )
        created.append(n[0])
        return n[0]

    schema_cache = {}

    def _find_schema(schema_parent, name, prefix):
        cache_key = (schema_parent.cdata, name, prefix)
        snode, module = schema_cache.get(cache_key, (None, None))
        if snode is not None:
            return snode, module
        if isinstance(schema_parent, SRpc):
            if rpcreply:
                schema_parent = schema_parent.output()
            else:
                schema_parent = schema_parent.input()

            if schema_parent is None:
                # there may not be any input or any output node in the rpc
                return None, None
        for s in schema_parent:
            if s.name() != name:
                continue
            mod = s.module()
            if prefix is not None and mod.name() != prefix:
                continue
            snode = s
            module = mod
            break
        schema_cache[cache_key] = (snode, module)
        return snode, module

    keys_cache = {}

    def _dic_keys(_dic, _schema):
        if isinstance(_schema, SList):
            # list keys must be first and in the order specified in the schema
            list_keys = keys_cache.get(_schema.cdata, None)
            if list_keys is None:
                list_keys = tuple(k.name() for k in _schema.keys())
                keys_cache[_schema.cdata] = list_keys
            keys = []
            for k in list_keys:
                if k in _dic:
                    keys.append(k)
            keys.extend(_dic.keys() - list_keys)
            return keys
        return _dic.keys()

    def _to_dnode(_dic, _schema, _parent=ffi.NULL, in_rpc_output=False):
        for key in _dic_keys(_dic, _schema):
            if ":" in key:
                prefix, name = key.split(":")
            else:
                prefix, name = None, key

            s, module = _find_schema(_schema, name, prefix)
            if not s:
                if isinstance(_schema, Module):
                    path = _schema.name()
                elif isinstance(_schema, SNode):
                    path = _schema.schema_path()
                else:
                    path = str(_schema)
                if strict:
                    raise LibyangError("%s: unknown element %r" % (path, key))
                LOG.warning("%s: skipping unknown element %r", path, key)
                continue

            value = _dic[key]

            if isinstance(s, SLeaf):
                _create_leaf(_parent, module, name, value, in_rpc_output)

            elif isinstance(s, SLeafList):
                if not isinstance(value, (list, tuple)):
                    raise TypeError(
                        "%s: python value is not a list/tuple: %r"
                        % (s.schema_path(), value)
                    )
                for v in value:
                    _create_leaf(_parent, module, name, v, in_rpc_output)

            elif isinstance(s, SRpc):
                n = _create_container(_parent, module, name, in_rpc_output)
                _to_dnode(value, s, n, rpcreply)

            elif isinstance(s, SContainer):
                n = _create_container(_parent, module, name, in_rpc_output)
                _to_dnode(value, s, n, in_rpc_output)

            elif isinstance(s, SList):
                if not isinstance(value, (list, tuple)):
                    raise TypeError(
                        "%s: python value is not a list/tuple: %r"
                        % (s.schema_path(), value)
                    )

                keys = [i.name() for i in s.keys()]
                for v in value:
                    if not isinstance(v, dict):
                        raise TypeError(
                            "%s: list element is not a dict: %r"
                            % (_schema.schema_path(), v)
                        )
                    val = v.copy()
                    key_values = []
                    for k in keys:
                        try:
                            key_values.append(val.pop(k))
                        except KeyError as e:
                            raise ValueError("Missing key %s in the list" % (k)) from e

                    n = _create_list(_parent, module, name, key_values, in_rpc_output)
                    _to_dnode(val, s, n, in_rpc_output)

            elif isinstance(s, SNotif):
                n = _create_container(_parent, module, name, in_rpc_output)
                _to_dnode(value, s, n, in_rpc_output)

    result = None

    try:
        if parent is not None:
            _parent = parent.cdata
            _schema_parent = parent.schema()
        else:
            _parent = ffi.NULL
            _schema_parent = module

        _to_dnode(
            dic,
            _schema_parent,
            _parent,
            in_rpc_output=rpcreply and isinstance(parent, DRpc),
        )
        if created:
            result = DNode.new(module.context, created[0])
            if validate:
                result.root().validate(
                    no_state=no_state,
                    validate_present=validate_present,
                    rpc=rpc,
                    rpcreply=rpcreply,
                    notification=notification,
                )

    except:
        for c in reversed(created):
            lib.lyd_free_tree(c)
        raise

    return result



# Copyright (c) 2018-2019 Robin Jarry
# SPDX-License-Identifier: MIT

import logging
import os
from typing import Any, Optional
import unittest

from libyang import (
    Context,
    ExtensionCompiled,
    ExtensionParsed,
    ExtensionPlugin,
    LibyangError,
    LibyangExtensionError,
    Module,
    PLeaf,
    SLeaf,
)


YANG_DIR = os.path.join(os.path.dirname(__file__), "yang")


# -------------------------------------------------------------------------------------
class TestExtensionPlugin(ExtensionPlugin):
    def __init__(self, context: Context) -> None:
        super().__init__(
            "omg-extensions",
            "type-desc",
            "omg-extensions-type-desc-plugin-v1",
            context,
            parse_clb=self._parse_clb,
            compile_clb=self._compile_clb,
            parse_free_clb=self._parse_free_clb,
            compile_free_clb=self._compile_free_clb,
        )
        self.parse_clb_called = 0
        self.compile_clb_called = 0
        self.parse_free_clb_called = 0
        self.compile_free_clb_called = 0
        self.parse_clb_exception: Optional[LibyangExtensionError] = None
        self.compile_clb_exception: Optional[LibyangExtensionError] = None
        self.parse_parent_stmt = None

    def reset(self) -> None:
        self.parse_clb_called = 0
        self.compile_clb_called = 0
        self.parse_free_clb_called = 0
        self.compile_free_clb_called = 0
        self.parse_clb_exception = None
        self.compile_clb_exception = None

    def _parse_clb(self, module: Module, ext: ExtensionParsed) -> None:
        self.parse_clb_called += 1
        if self.parse_clb_exception is not None:
            raise self.parse_clb_exception
        self.parse_substmts(ext)
        self.parse_parent_stmt = self.stmt2str(ext.cdata.parent_stmt)

    def _compile_clb(self, pext: ExtensionParsed, cext: ExtensionCompiled) -> None:
        self.compile_clb_called += 1
        if self.compile_clb_exception is not None:
            raise self.compile_clb_exception
        self.compile_substmts(pext, cext)

    def _parse_free_clb(self, ext: ExtensionParsed) -> None:
        self.parse_free_clb_called += 1
        self.free_parse_substmts(ext)

    def _compile_free_clb(self, ext: ExtensionCompiled) -> None:
        self.compile_free_clb_called += 1
        self.free_compile_substmts(ext)


# -------------------------------------------------------------------------------------
class ExtensionTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR)
        self.plugin = TestExtensionPlugin(self.ctx)

    def tearDown(self):
        self.ctx.destroy()
        self.ctx = None

    def test_extension_basic(self):
        self.ctx.load_module("yolo-system")
        self.assertEqual(5, self.plugin.parse_clb_called)
        self.assertEqual(6, self.plugin.compile_clb_called)
        self.assertEqual(0, self.plugin.parse_free_clb_called)
        self.assertEqual(0, self.plugin.compile_free_clb_called)
        self.assertEqual("type", self.plugin.parse_parent_stmt)
        self.ctx.destroy()
        self.assertEqual(5, self.plugin.parse_clb_called)
        self.assertEqual(6, self.plugin.compile_clb_called)
        self.assertEqual(5, self.plugin.parse_free_clb_called)
        self.assertEqual(6, self.plugin.compile_free_clb_called)

    def test_extension_invalid_parse(self):
        self.plugin.parse_clb_exception = LibyangExtensionError(
            "this extension cannot be parsed",
            self.plugin.ERROR_NOT_VALID,
            logging.ERROR,
        )
        with self.assertRaises(LibyangError):
            self.ctx.load_module("yolo-system")

    def test_extension_invalid_compile(self):
        self.plugin.compile_clb_exception = LibyangExtensionError(
            "this extension cannot be compiled",
            self.plugin.ERROR_NOT_VALID,
            logging.ERROR,
        )
        with self.assertRaises(LibyangError):
            self.ctx.load_module("yolo-system")


# -------------------------------------------------------------------------------------
class ExampleParseExtensionPlugin(ExtensionPlugin):
    def __init__(self, context: Context) -> None:
        super().__init__(
            "omg-extensions",
            "parse-validation",
            "omg-extensions-parse-validation-plugin-v1",
            context,
            parse_clb=self._parse_clb,
        )

    def _verify_single(self, parent: Any) -> None:
        count = 0
        for e in parent.extensions():
            if e.name() == self.name and e.module().name() == self.module_name:
                count += 1
            if count > 1:
                raise LibyangExtensionError(
                    f"Extension {self.name} is allowed to be defined just once per given "
                    "parent node context.",
                    self.ERROR_NOT_VALID,
                    logging.ERROR,
                )

    def _parse_clb(self, _, ext: ExtensionParsed) -> None:
        parent = ext.parent_node()
        if not isinstance(parent, PLeaf):
            raise LibyangExtensionError(
                f"Extension {ext.name()} is allowed only in leaf nodes",
                self.ERROR_NOT_VALID,
                logging.ERROR,
            )
        self._verify_single(parent)
        # here you put code to perform something reasonable actions you need for your extension


class ExampleCompileExtensionPlugin(ExtensionPlugin):
    def __init__(self, context: Context) -> None:
        super().__init__(
            "omg-extensions",
            "compile-validation",
            "omg-extensions-compile-validation-plugin-v1",
            context,
            compile_clb=self._compile_clb,
        )

    def _compile_clb(self, pext: ExtensionParsed, cext: ExtensionCompiled) -> None:
        parent = cext.parent_node()
        if not isinstance(parent, SLeaf):
            raise LibyangExtensionError(
                f"Extension {cext.name()} is allowed only in leaf nodes",
                self.ERROR_NOT_VALID,
                logging.ERROR,
            )
        # here you put code to perform something reasonable actions you need for your extension


class ExtensionExampleTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR)
        self.plugins = []

    def tearDown(self):
        self.plugins.clear()
        self.ctx.destroy()
        self.ctx = None

    def test_parse_validation_example(self):
        self.plugins.append(ExampleParseExtensionPlugin(self.ctx))
        self.ctx.load_module("yolo-system")

    def test_compile_validation_example(self):
        self.plugins.append(ExampleParseExtensionPlugin(self.ctx))
        self.plugins.append(ExampleCompileExtensionPlugin(self.ctx))
        with self.assertRaises(LibyangError):
            self.ctx.load_module("yolo-system")



# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: MIT

import unittest

import libyang as ly


# -------------------------------------------------------------------------------------
class KeyedListTest(unittest.TestCase):
    def test_keylist_base(self):
        l = ly.KeyedList(["x"])
        with self.assertRaises(TypeError):
            l.index("x")
        with self.assertRaises(TypeError):
            l.insert(0, "x")
        with self.assertRaises(TypeError):
            l.reverse()
        with self.assertRaises(TypeError):
            l.sort()
        with self.assertRaises(TypeError):
            l["x"] = 2
        with self.assertRaises(TypeError):
            l[0]  # pylint: disable=pointless-statement
        with self.assertRaises(TypeError):
            del l[0]
        with self.assertRaises(TypeError):
            l.pop()
        with self.assertRaises(TypeError):
            l.pop(0)

    def test_keylist_leaflist(self):
        l = ly.KeyedList("abcdef")
        self.assertEqual(len(l), 6)
        self.assertIn("e", l)
        l.remove("e")
        self.assertNotIn("e", l)
        self.assertIn("c", l)
        del l["c"]
        self.assertNotIn("c", l)
        with self.assertRaises(KeyError):
            del l["x"]
        with self.assertRaises(KeyError):
            l.remove("x")
        self.assertEqual(l.pop("x", "not-found"), "not-found")
        self.assertEqual(l.pop("b"), "b")
        self.assertNotIn("b", l)
        l.clear()
        self.assertEqual(len(l), 0)
        self.assertEqual(l, [])

    def test_keylist_list_one_key(self):
        l = ly.KeyedList(
            [
                {"mtu": 1500, "name": "eth0"},
                {"mtu": 512, "name": "eth2"},
                {"mtu": 1280, "name": "eth4"},
            ],
            key_name="name",
        )
        self.assertIn("eth2", l)
        self.assertEqual(l["eth4"], {"mtu": 1280, "name": "eth4"})
        del l["eth0"]
        self.assertEqual(
            l, [{"mtu": 512, "name": "eth2"}, {"mtu": 1280, "name": "eth4"}]
        )
        l.append({"name": "eth6", "mtu": 6})
        l.append({"name": "eth5", "mtu": 5})
        l.append({"name": "eth10", "mtu": 10})
        self.assertEqual(
            l,
            # order does not matter
            [
                {"mtu": 10, "name": "eth10"},
                {"mtu": 512, "name": "eth2"},
                {"mtu": 5, "name": "eth5"},
                {"mtu": 6, "name": "eth6"},
                {"mtu": 1280, "name": "eth4"},
            ],
        )
        self.assertEqual(l.pop("eth10000", {}), {})
        self.assertEqual(l.pop("eth10"), {"mtu": 10, "name": "eth10"})
        self.assertNotIn("eth10", l)
        l.clear()
        self.assertEqual(len(l), 0)
        self.assertEqual(l, [])

    def test_keylist_list_multiple_keys(self):
        l = ly.KeyedList(
            [
                {"id": 5, "size": 10, "desc": "foo"},
                {"id": 8, "size": 5, "desc": "bar"},
                {"id": 2, "size": 14, "desc": "baz"},
            ],
            key_name=("id", "size"),
        )
        self.assertIn(("8", "5"), l)
        self.assertEqual(l["8", "5"], {"id": 8, "size": 5, "desc": "bar"})
        self.assertEqual(l[("2", "14")], {"id": 2, "size": 14, "desc": "baz"})
        del l["2", "14"]
        self.assertEqual(
            l,
            [{"id": 5, "size": 10, "desc": "foo"}, {"id": 8, "size": 5, "desc": "bar"}],
        )
        l.clear()
        self.assertEqual(len(l), 0)
        self.assertEqual(l, [])



# Copyright (c) 2018-2019 Robin Jarry
# SPDX-License-Identifier: MIT

import os
import unittest

from libyang import (
    Context,
    Extension,
    ExtensionCompiled,
    ExtensionParsed,
    Identity,
    IfFeature,
    IfOrFeatures,
    IOType,
    LibyangError,
    Module,
    Must,
    PAction,
    PActionInOut,
    PAnydata,
    Pattern,
    PAugment,
    PCase,
    PChoice,
    PContainer,
    PGrouping,
    PIdentity,
    PLeaf,
    PLeafList,
    PList,
    PNode,
    PNotif,
    PRefine,
    PType,
    PUses,
    Revision,
    SAnydata,
    SCase,
    SChoice,
    SContainer,
    SLeaf,
    SLeafList,
    SList,
    SNode,
    SRpc,
    Type,
)


YANG_DIR = os.path.join(os.path.dirname(__file__), "yang")


# -------------------------------------------------------------------------------------
class ModuleTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR)
        self.module = self.ctx.load_module("yolo-system")

    def tearDown(self):
        self.module = None
        self.ctx.destroy()
        self.ctx = None

    def test_mod_print_mem(self):
        s = self.module.print("tree", IOType.MEMORY)
        self.assertGreater(len(s), 0)

        s = self.module.print_mem("tree")
        self.assertGreater(len(s), 0)

    def test_mod_attrs(self):
        self.assertEqual(self.module.name(), "yolo-system")
        self.assertEqual(self.module.description(), "YOLO.")
        self.assertEqual(self.module.prefix(), "sys")

    def test_mod_filepath(self):
        self.assertEqual(
            self.module.filepath(), os.path.join(YANG_DIR, "yolo/yolo-system.yang")
        )

    def test_mod_iter(self):
        children = list(iter(self.module))
        self.assertEqual(len(children), 6)

    def test_mod_children_rpcs(self):
        rpcs = list(self.module.children(types=(SNode.RPC,)))
        self.assertEqual(len(rpcs), 2)

    def test_mod_enable_features(self):
        self.assertFalse(self.module.feature_state("turbo-boost"))
        self.module.feature_enable("turbo-boost")
        self.module.feature_enable("*")
        self.assertTrue(self.module.feature_state("turbo-boost"))
        self.module.feature_disable_all()
        self.assertFalse(self.module.feature_state("turbo-boost"))
        self.module.feature_enable_all()
        self.assertTrue(self.module.feature_state("turbo-boost"))
        self.module.feature_disable_all()

    def test_mod_imports(self):
        imports = list(self.module.imports())
        self.assertEqual(imports[0].name(), "omg-extensions")
        self.assertEqual(imports[1].name(), "wtf-types")
        self.assertEqual(len(imports), 2)

    def test_mod_features(self):
        features = list(self.module.features())
        self.assertEqual(len(features), 2)

    def test_mod_get_feature(self):
        self.module.feature_enable("turbo-boost")
        feature = self.module.get_feature("turbo-boost")
        self.assertEqual(feature.name(), "turbo-boost")
        self.assertEqual(feature.description(), "Goes faster.")
        self.assertIsNone(feature.reference())
        self.assertTrue(feature.state())
        self.assertFalse(feature.deprecated())
        self.assertFalse(feature.obsolete())

    def test_mod_get_feature_not_found(self):
        with self.assertRaises(LibyangError):
            self.module.get_feature("does-not-exist")

    def test_mod_revisions(self):
        revisions = list(self.module.revisions())
        self.assertEqual(len(revisions), 2)
        self.assertIsInstance(revisions[0], Revision)
        self.assertEqual(revisions[0].date(), "1999-04-01")
        self.assertEqual(revisions[1].date(), "1990-04-01")

    def test_mod_extensions(self):
        assert self.module is not None  # pyright doesn't understand assertIsNotNone()
        exts = list(self.module.extensions())
        self.assertEqual(len(exts), 1)
        ext = self.module.get_extension("compile-validation", prefix="omg-extensions")
        self.assertEqual(ext.argument(), "module-level")
        sub_exts = list(ext.extensions())
        self.assertEqual(len(sub_exts), 1)
        ext = sub_exts[0]
        self.assertEqual(ext.name(), "compile-validation")
        self.assertEqual(ext.module().name(), "omg-extensions")
        self.assertEqual(ext.argument(), "module-sub-level")


# -------------------------------------------------------------------------------------
class RevisionTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR)
        mod = self.ctx.load_module("yolo-system")
        revisions = list(mod.revisions())
        self.revision = revisions[0]

    def tearDown(self):
        self.revision = None
        self.ctx.destroy()
        self.ctx = None

    def test_rev_date(self):
        self.assertEqual(self.revision.date(), "1999-04-01")

    def test_rev_reference(self):
        self.assertEqual(
            self.revision.reference(),
            "RFC 2549 - IP over Avian Carriers with Quality of Service.",
        )

    def test_rev_description(self):
        self.assertEqual(self.revision.description(), "Version update.")

    def test_rev_extensions(self):
        exts = list(self.revision.extensions())
        self.assertEqual(len(exts), 1)
        ext = self.revision.get_extension("human-name", prefix="omg-extensions")
        self.assertIsInstance(ext, Extension)


# -------------------------------------------------------------------------------------
class IfFeatureTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR)
        self.mod = self.ctx.load_module("yolo-system")
        self.mod.feature_enable_all()
        self.leaf = next(
            self.ctx.find_path("/yolo-system:conf/yolo-system:isolation-level")
        )

    def tearDown(self):
        self.ctx.destroy()
        self.ctx = None

    def test_iffeatures(self):
        iffeatures = list(self.leaf.if_features())
        self.assertEqual(len(iffeatures), 1)

    def test_iffeature_tree(self):
        iff = next(self.leaf.if_features())
        tree = iff.tree()
        self.assertIsInstance(tree, IfOrFeatures)
        self.assertIsInstance(tree.a, IfFeature)
        self.assertIsInstance(tree.b, IfFeature)
        self.assertEqual(tree.a.feature().name(), "turbo-boost")
        self.assertEqual(tree.b.feature().name(), "networking")

    def test_iffeature_state(self):
        def feature_disable_only(feature):
            self.mod.feature_disable_all()
            for f in self.mod.features():
                if f.name() == feature:
                    continue
                self.mod.feature_enable(f.name())

        # if-feature is just a feature
        leaf_simple = next(self.ctx.find_path("/yolo-system:conf/yolo-system:speed"))
        tree = next(leaf_simple.if_features()).tree()
        self.mod.feature_enable_all()
        self.assertEqual(tree.state(), True)
        self.mod.feature_disable_all()
        self.assertEqual(tree.state(), False)

        # if-feature is "NOT networking"
        self.mod.feature_disable_all()
        leaf_not = next(self.ctx.find_path("/yolo-system:conf/yolo-system:offline"))
        tree = next(leaf_not.if_features()).tree()
        self.mod.feature_enable_all()
        self.assertEqual(tree.state(), False)
        self.mod.feature_disable_all()
        self.assertEqual(tree.state(), True)

        # if-feature is "turbo-boost AND networking"
        self.mod.feature_enable_all()
        leaf_and = next(self.ctx.find_path("/yolo-system:conf/yolo-system:full"))
        tree = next(leaf_and.if_features()).tree()
        self.mod.feature_enable_all()
        self.assertEqual(tree.state(), True)
        self.mod.feature_disable_all()
        self.assertEqual(tree.state(), False)
        feature_disable_only("turbo-boost")
        self.assertEqual(tree.state(), False)
        feature_disable_only("networking")
        self.assertEqual(tree.state(), False)

        # if-feature is "turbo-boost OR networking"
        leaf_or = next(
            self.ctx.find_path("/yolo-system:conf/yolo-system:isolation-level")
        )
        tree = next(leaf_or.if_features()).tree()
        self.mod.feature_enable_all()
        self.assertEqual(tree.state(), True)
        self.mod.feature_disable_all()
        self.assertEqual(tree.state(), False)
        feature_disable_only("turbo-boost")
        self.assertEqual(tree.state(), True)
        feature_disable_only("networking")
        self.assertEqual(tree.state(), True)

    def test_iffeature_str(self):
        iff = next(self.leaf.if_features())
        self.assertEqual(str(iff), "turbo-boost OR networking")

    def test_iffeature_dump(self):
        iff = next(self.leaf.if_features())
        self.assertEqual(
            iff.dump(),
            """OR
 turbo-boost [Goes faster.]
 networking [Supports networking.]
""",
        )


# -------------------------------------------------------------------------------------
class ContainerTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR, compile_obsolete=True)
        mod = self.ctx.load_module("yolo-system")
        mod.feature_enable_all()
        self.container = next(self.ctx.find_path("/yolo-system:conf"))

    def tearDown(self):
        self.container = None
        self.ctx.destroy()
        self.ctx = None

    def test_cont_attrs(self):
        self.assertIsInstance(self.container, SContainer)
        self.assertEqual(self.container.nodetype(), SNode.CONTAINER)
        self.assertEqual(self.container.keyword(), "container")
        self.assertEqual(self.container.name(), "conf")
        self.assertEqual(self.container.fullname(), "yolo-system:conf")
        self.assertEqual(self.container.description(), "Configuration.")
        self.assertEqual(self.container.config_set(), False)
        self.assertEqual(self.container.config_false(), False)
        self.assertEqual(self.container.mandatory(), False)
        self.assertIsInstance(self.container.module(), Module)
        self.assertEqual(self.container.schema_path(), "/yolo-system:conf")
        self.assertEqual(self.container.data_path(), "/yolo-system:conf")
        self.assertIs(self.container.presence(), None)

    def test_cont_iter(self):
        children = list(iter(self.container))
        self.assertEqual(len(children), 11)

    def test_cont_children_leafs(self):
        leafs = list(self.container.children(types=(SNode.LEAF,)))
        self.assertEqual(len(leafs), 9)
        without_choice = [c.name() for c in self.container.children(with_choice=False)]
        with_choice = [c.name() for c in self.container.children(with_choice=True)]
        self.assertTrue("pill" not in without_choice)
        self.assertTrue("pill" in with_choice)

    def test_cont_parent(self):
        self.assertIsNone(self.container.parent())

    def test_iter_tree(self):
        tree = list(self.container.iter_tree())
        self.assertEqual(len(tree), 20)
        tree = list(self.container.iter_tree(full=True))
        self.assertEqual(len(tree), 25)

    def test_container_parsed(self):
        pnode = self.container.parsed()
        self.assertIsInstance(pnode, PContainer)
        self.assertIsNone(next(pnode.musts(), None))
        self.assertIsNone(pnode.when_condition())
        self.assertIsNone(pnode.presence())
        self.assertIsNone(next(pnode.typedefs(), None))
        self.assertIsNone(next(pnode.groupings(), None))
        self.assertIsNotNone(next(iter(pnode)))
        self.assertIsNone(next(pnode.actions(), None))
        self.assertIsNone(next(pnode.notifications(), None))


# -------------------------------------------------------------------------------------
class UsesTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR)
        mod = self.ctx.load_module("yolo-nodetypes")
        mod.feature_enable_all()

    def tearDown(self):
        self.ctx.destroy()
        self.ctx = None

    def test_uses_parsed(self):
        snode = next(self.ctx.find_path("/yolo-nodetypes:cont2"))
        self.assertIsInstance(snode, SContainer)
        pnode = snode.parsed()
        self.assertIsInstance(pnode, PContainer)
        pnode = next(iter(pnode))
        self.assertIsInstance(pnode, PUses)

        ref_pnode = next(pnode.refines())
        self.assertIsInstance(ref_pnode, PRefine)
        self.assertEqual("cont3/leaf1", ref_pnode.nodeid())
        self.assertIsNone(ref_pnode.description())
        self.assertIsNone(ref_pnode.reference())
        self.assertIsNone(next(ref_pnode.if_features(), None))
        self.assertIsNone(next(ref_pnode.musts(), None))
        self.assertIsNone(ref_pnode.presence())
        self.assertIsNone(next(ref_pnode.defaults(), None))
        self.assertEqual(0, ref_pnode.min_elements())
        self.assertIsNone(ref_pnode.max_elements())
        self.assertIsNone(next(ref_pnode.extensions(), None))

        aug_pnode = next(pnode.augments())
        self.assertIsInstance(aug_pnode, PAugment)
        self.assertIsNotNone(next(iter(aug_pnode)))
        self.assertIsNone(aug_pnode.when_condition())
        self.assertIsNone(next(aug_pnode.actions(), None))
        self.assertIsNone(next(aug_pnode.notifications(), None))


# -------------------------------------------------------------------------------------
class GroupingTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR)

    def tearDown(self):
        self.ctx.destroy()
        self.ctx = None

    def test_grouping_parsed(self):
        mod = self.ctx.load_module("yolo-nodetypes")
        pnode = next(mod.groupings())
        self.assertIsInstance(pnode, PGrouping)
        self.assertIsNone(next(pnode.typedefs(), None))
        self.assertIsNone(next(pnode.groupings(), None))
        child = next(iter(pnode))
        self.assertIsNotNone(child)
        self.assertIsNone(next(pnode.actions(), None))
        self.assertIsNone(next(pnode.notifications(), None))


# -------------------------------------------------------------------------------------
class ListTest(unittest.TestCase):
    PATH = {
        "LOG": "/yolo-system:conf/url",
        "DATA": "/yolo-system:conf/url",
        "DATA_PATTERN": "/yolo-system:conf/url[proto='%s'][host='%s']",
    }

    def setUp(self):
        self.ctx = Context(YANG_DIR)
        self.ctx.load_module("yolo-system")
        self.ctx.load_module("yolo-nodetypes")
        self.list = next(self.ctx.find_path(self.PATH["LOG"]))

    def tearDown(self):
        self.list = None
        self.ctx.destroy()
        self.ctx = None

    def test_list_attrs(self):
        self.assertIsInstance(self.list, SList)
        self.assertEqual(self.list.nodetype(), SNode.LIST)
        self.assertEqual(self.list.keyword(), "list")

        self.assertEqual(self.list.schema_path(), self.PATH["LOG"])

        self.assertEqual(self.list.schema_path(SNode.PATH_DATA), self.PATH["DATA"])

        self.assertEqual(self.list.data_path(), self.PATH["DATA_PATTERN"])
        self.assertFalse(self.list.ordered())

    def test_list_keys(self):
        keys = list(self.list.keys())
        self.assertEqual(len(keys), 2)

    def test_list_iter(self):
        children = list(iter(self.list))
        self.assertEqual(len(children), 6)

    def test_list_children_skip_keys(self):
        children = list(self.list.children(skip_keys=True))
        self.assertEqual(len(children), 4)

    def test_list_parent(self):
        parent = self.list.parent()
        self.assertIsNotNone(parent)
        self.assertIsInstance(parent, SContainer)
        self.assertEqual(parent.name(), "conf")

    def test_list_uniques(self):
        list1 = next(self.ctx.find_path("/yolo-nodetypes:conf/list1"))
        self.assertIsInstance(list1, SList)
        uniques = list(list1.uniques())
        self.assertEqual(len(uniques), 1)
        elements = [u.name() for u in uniques[0]]
        self.assertEqual(len(elements), 2)
        self.assertTrue("leaf2" in elements)
        self.assertTrue("leaf3" in elements)

        list2 = next(self.ctx.find_path("/yolo-nodetypes:conf/list2"))
        self.assertIsInstance(list2, SList)
        uniques = list(list2.uniques())
        self.assertEqual(len(uniques), 0)

    def test_list_min_max(self):
        list1 = next(self.ctx.find_path("/yolo-nodetypes:conf/list1"))
        self.assertIsInstance(list1, SList)
        self.assertEqual(list1.min_elements(), 2)
        self.assertEqual(list1.max_elements(), 10)

        list2 = next(self.ctx.find_path("/yolo-nodetypes:conf/list2"))
        self.assertIsInstance(list2, SList)
        self.assertEqual(list2.min_elements(), 0)
        self.assertEqual(list2.max_elements(), None)

    def test_list_parsed(self):
        list1 = next(self.ctx.find_path("/yolo-nodetypes:conf/list1"))
        self.assertIsInstance(list1, SList)
        pnode = list1.parsed()
        self.assertIsInstance(pnode, PList)
        self.assertIsNone(next(pnode.musts(), None))
        self.assertIsNone(pnode.when_condition())
        self.assertEqual("leaf1", pnode.key())
        self.assertIsNone(next(pnode.typedefs(), None))
        self.assertIsNone(next(pnode.groupings(), None))
        child = next(iter(pnode))
        self.assertIsInstance(child, PLeaf)
        self.assertIsNone(next(pnode.actions(), None))
        self.assertIsNone(next(pnode.notifications(), None))
        self.assertEqual("leaf2 leaf3", next(pnode.uniques()))
        self.assertEqual(2, pnode.min_elements())
        self.assertEqual(10, pnode.max_elements())
        self.assertFalse(pnode.ordered())


# -------------------------------------------------------------------------------------
class RpcTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR)
        self.ctx.load_module("yolo-system")
        self.rpc = next(self.ctx.find_path("/yolo-system:format-disk"))

    def tearDown(self):
        self.rpc = None
        self.ctx.destroy()
        self.ctx = None

    def test_rpc_attrs(self):
        self.assertIsInstance(self.rpc, SRpc)
        self.assertEqual(self.rpc.nodetype(), SNode.RPC)
        self.assertEqual(self.rpc.keyword(), "rpc")
        self.assertEqual(self.rpc.schema_path(), "/yolo-system:format-disk")
        choice = next(self.rpc.input().children((SNode.CHOICE,), with_choice=True))
        self.assertIsInstance(choice, SChoice)

    def test_rpc_extensions(self):
        ext = list(self.rpc.extensions())
        self.assertEqual(len(ext), 1)
        ext = self.rpc.get_extension("require-admin", prefix="omg-extensions")
        self.assertIsInstance(ext, ExtensionCompiled)
        self.assertIsInstance(ext.parent_node(), SRpc)
        self.assertIsNone(next(ext.extensions(), None))
        parsed = self.rpc.parsed()
        ext = parsed.get_extension("require-admin", prefix="omg-extensions")
        self.assertIsInstance(ext, ExtensionParsed)
        self.assertIsInstance(ext.parent_node(), PAction)
        self.assertIsNone(next(ext.extensions(), None))

    def test_rpc_params(self):
        leaf = next(self.rpc.children())
        self.assertIsInstance(leaf, SLeaf)
        self.assertEqual(leaf.data_path(), "/yolo-system:format-disk/disk")
        leaf = next(self.rpc.input().children())
        self.assertIsInstance(leaf, SLeaf)

    def test_rpc_no_parent(self):
        self.assertIsNone(self.rpc.parent())

    def test_rpc_parsed(self):
        self.assertIsInstance(self.rpc, SRpc)
        pnode = self.rpc.parsed()
        self.assertIsInstance(pnode, PAction)
        self.assertIsNone(next(pnode.typedefs(), None))
        self.assertIsNone(next(pnode.groupings(), None))
        pnode2 = pnode.input()
        self.assertIsInstance(pnode2, PActionInOut)
        self.assertIsInstance(pnode.output(), PActionInOut)
        self.assertIsNone(next(pnode2.musts(), None))
        self.assertIsNone(next(pnode2.typedefs(), None))
        self.assertIsNone(next(pnode2.groupings(), None))
        pnode3 = next(iter(pnode2))
        self.assertIsInstance(pnode3, PLeaf)


# -------------------------------------------------------------------------------------
class LeafTypeTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR, compile_obsolete=True)
        self.ctx.load_module("yolo-system")

    def tearDown(self):
        self.ctx.destroy()
        self.ctx = None

    def test_leaf_type_derived(self):
        leaf = next(self.ctx.find_path("/yolo-system:conf/yolo-system:hostname"))
        self.assertIsInstance(leaf, SLeaf)
        t = leaf.type()
        self.assertIsInstance(t, Type)
        self.assertEqual(t.name(), "types:host")
        self.assertEqual(t.base(), Type.STRING)
        mod = t.module()
        self.assertIsNot(mod, None)
        self.assertEqual(mod.name(), "yolo-system")
        self.assertEqual(mod.get_module_from_prefix("types").name(), "wtf-types")
        self.assertEqual(t.typedef().name(), "host")
        self.assertEqual(t.typedef().description(), "my host type.")
        self.assertEqual(t.description(), "my host type.")

    def test_leaf_type_status(self):
        leaf = next(self.ctx.find_path("/yolo-system:conf/yolo-system:hostname"))
        self.assertIsInstance(leaf, SLeaf)
        self.assertEqual(leaf.deprecated(), False)
        self.assertEqual(leaf.obsolete(), False)
        leaf = next(self.ctx.find_path("/yolo-system:conf/yolo-system:deprecated-leaf"))
        self.assertIsInstance(leaf, SLeaf)
        self.assertEqual(leaf.deprecated(), True)
        self.assertEqual(leaf.obsolete(), False)
        leaf = next(self.ctx.find_path("/yolo-system:conf/yolo-system:obsolete-leaf"))
        self.assertIsInstance(leaf, SLeaf)
        self.assertEqual(leaf.deprecated(), False)
        self.assertEqual(leaf.obsolete(), True)

    def test_leaf_type_pattern(self):
        leaf = next(
            self.ctx.find_path("/yolo-system:conf/yolo-system:url/yolo-system:host")
        )
        self.assertIsInstance(leaf, SLeaf)
        t = leaf.type()
        self.assertIsInstance(t, Type)
        self.assertEqual(list(t.patterns()), [("[a-z.]+", False), ("1", True)])
        patterns = list(t.all_pattern_details())
        self.assertEqual(len(patterns), 2)
        self.assertIsInstance(patterns[0], Pattern)
        self.assertEqual(patterns[0].expression(), "[a-z.]+")
        self.assertFalse(patterns[0].inverted())
        self.assertEqual(patterns[0].error_message(), "ERROR1")
        self.assertEqual(patterns[1].expression(), "1")
        self.assertTrue(patterns[1].inverted())
        self.assertIsNone(patterns[1].error_message())

    def test_leaf_type_union(self):
        leaf = next(self.ctx.find_path("/yolo-system:conf/yolo-system:number"))
        self.assertIsInstance(leaf, SLeafList)
        t = leaf.type()
        self.assertIsInstance(t, Type)
        self.assertEqual(t.name(), "types:number")
        self.assertEqual(t.base(), Type.UNION)
        types = set(u.name() for u in t.union_types())
        types2 = set(u.name() for u in t.union_types(with_typedefs=True))
        self.assertEqual(types, set(["int16", "int32", "uint16", "uint32"]))
        self.assertEqual(types2, set(["signed", "unsigned"]))
        for u in t.union_types():
            ext = u.get_extension(
                "type-desc", prefix="omg-extensions", arg_value=f"<{u.name()}>"
            )
            self.assertIsInstance(ext, Extension)
            self.assertEqual(len(list(u.extensions())), 2)
        bases = set(t.basenames())
        self.assertEqual(bases, set(["int16", "int32", "uint16", "uint32"]))

    def test_leaf_type_extensions(self):
        leaf = next(
            self.ctx.find_path("/yolo-system:conf/yolo-system:url/yolo-system:proto")
        )
        t = leaf.type()
        ext = t.get_extension(
            "type-desc", prefix="omg-extensions", arg_value="<protocol>"
        )
        self.assertIsInstance(ext, Extension)
        self.assertIsNone(ext.parent_node())

    def test_leaf_type_enum(self):
        leaf = next(
            self.ctx.find_path("/yolo-system:conf/yolo-system:url/yolo-system:proto")
        )
        self.assertIsInstance(leaf, SLeaf)
        t = leaf.type()
        self.assertIsInstance(t, Type)
        self.assertEqual(t.name(), "types:protocol")
        self.assertEqual(t.base(), Type.ENUM)
        enums = [e.name() for e in t.enums()]
        self.assertEqual(enums, ["http", "https", "ftp", "sftp"])
        enum = next(t.enums())
        self.assertIsNone(next(enum.extensions(), None))
        self.assertIsNone(enum.get_extension("test", prefix="test"))

    def test_leaf_type_bits(self):
        leaf = next(self.ctx.find_path("/yolo-system:chmod/yolo-system:perms"))
        self.assertIsInstance(leaf, SLeaf)
        t = leaf.type()
        self.assertIsInstance(t, Type)
        self.assertEqual(t.name(), "types:permissions")
        self.assertEqual(t.base(), Type.BITS)
        bits = [b.name() for b in t.bits()]
        self.assertEqual(bits, ["read", "write", "execute"])

    def test_leaf_parent(self):
        leaf = next(
            self.ctx.find_path("/yolo-system:conf/yolo-system:url/yolo-system:proto")
        )
        parent = leaf.parent()
        self.assertIsNotNone(parent)
        self.assertIsInstance(parent, SList)
        self.assertEqual(parent.name(), "url")

    def test_iter_tree(self):
        leaf = next(self.ctx.find_path("/yolo-system:conf"))
        self.assertEqual(len(list(leaf.iter_tree(full=True))), 23)

    def test_leaf_type_fraction_digits(self):
        self.ctx.load_module("yolo-nodetypes")
        leaf = next(self.ctx.find_path("/yolo-nodetypes:conf/percentage"))
        self.assertIsInstance(leaf, SLeaf)
        t = leaf.type()
        self.assertIsInstance(t, Type)
        self.assertEqual(next(t.all_fraction_digits(), None), 2)

    def test_leaf_type_require_instance(self):
        leaf = next(self.ctx.find_path("/yolo-system:conf/hostname-ref"))
        self.assertIsInstance(leaf, SLeaf)
        t = leaf.type()
        self.assertIsInstance(t, Type)
        self.assertFalse(t.require_instance())

    def test_leaf_type_parsed(self):
        leaf = next(self.ctx.find_path("/yolo-system:conf/yolo-system:hostname"))
        self.assertIsInstance(leaf, SLeaf)
        t = leaf.type()
        self.assertIsInstance(t, Type)
        pnode = t.parsed()
        self.assertIsInstance(pnode, PType)
        self.assertEqual("types:host", pnode.name())
        self.assertIsNone(pnode.range())
        self.assertIsNone(pnode.length())
        self.assertIsNone(next(pnode.patterns(), None))
        self.assertIsNone(next(pnode.enums(), None))
        self.assertIsNone(next(pnode.bits(), None))
        self.assertIsNone(pnode.path())
        self.assertIsNone(next(pnode.bases(), None))
        self.assertIsNone(next(pnode.types(), None))
        self.assertIsNone(next(pnode.extensions(), None))
        self.assertIsNotNone(pnode.pmod())
        self.assertIsNone(pnode.compiled())
        self.assertEqual(0, pnode.fraction_digits())
        self.assertFalse(pnode.require_instance())


# -------------------------------------------------------------------------------------
class LeafTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR)
        self.ctx.load_module("yolo-nodetypes")

    def tearDown(self):
        self.ctx.destroy()
        self.ctx = None

    def test_must(self):
        leaf = next(self.ctx.find_path("/yolo-nodetypes:conf/percentage"))
        self.assertIsInstance(leaf, SLeaf)
        must = next(leaf.musts(), None)
        self.assertIsInstance(must, Must)
        self.assertEqual(must.error_message(), "ERROR1")
        must = next(leaf.must_conditions(), None)
        self.assertIsInstance(must, str)

    def test_leaf_default(self):
        leaf = next(self.ctx.find_path("/yolo-nodetypes:conf/percentage"))
        self.assertIsInstance(leaf.default(), float)

    def test_leaf_parsed(self):
        leaf = next(self.ctx.find_path("/yolo-nodetypes:conf/percentage"))
        self.assertIsInstance(leaf, SLeaf)
        pnode = leaf.parsed()
        self.assertIsInstance(pnode, PLeaf)
        must = next(pnode.musts())
        self.assertIsInstance(must, Must)
        self.assertEqual(must.error_message(), "ERROR1")
        must = next(leaf.must_conditions())
        self.assertIsInstance(must, str)
        self.assertIsNone(pnode.when_condition())
        self.assertIsInstance(pnode.type(), PType)
        self.assertIsNone(pnode.units())
        self.assertEqual("10.2", pnode.default())
        self.assertFalse(pnode.is_key())

        # test basic PNode settings
        self.assertIsNotNone(pnode.parent())
        self.assertEqual(PNode.LEAF, pnode.nodetype())
        self.assertIsNotNone(next(pnode.siblings()))
        self.assertEqual("<libyang.schema.PLeaf: percentage>", repr(pnode))
        self.assertIsNone(pnode.description())
        self.assertIsNone(pnode.reference())
        self.assertIsNone(next(pnode.if_features(), None))
        self.assertIsNone(next(pnode.extensions(), None))
        self.assertIsNone(pnode.get_extension("test", prefix="test"))
        self.assertFalse(pnode.config_set())
        self.assertFalse(pnode.config_false())
        self.assertFalse(pnode.mandatory())
        self.assertFalse(pnode.deprecated())
        self.assertFalse(pnode.obsolete())
        self.assertEqual("current", pnode.status())

    NODETYPE_CLASS = {}


# -------------------------------------------------------------------------------------
class LeafListTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR)
        self.ctx.load_module("yolo-nodetypes")

    def tearDown(self):
        self.ctx.destroy()
        self.ctx = None

    def test_leaflist_defaults(self):
        leaflist = next(self.ctx.find_path("/yolo-nodetypes:conf/ratios"))
        for d in leaflist.defaults():
            self.assertIsInstance(d, float)
        leaflist = next(self.ctx.find_path("/yolo-nodetypes:conf/bools"))
        for d in leaflist.defaults():
            self.assertIsInstance(d, bool)
        leaflist = next(self.ctx.find_path("/yolo-nodetypes:conf/integers"))
        for d in leaflist.defaults():
            self.assertIsInstance(d, int)

    def test_leaf_list_min_max(self):
        leaflist1 = next(self.ctx.find_path("/yolo-nodetypes:conf/leaf-list1"))
        self.assertIsInstance(leaflist1, SLeafList)
        self.assertEqual(leaflist1.min_elements(), 3)
        self.assertEqual(leaflist1.max_elements(), 11)

        leaflist2 = next(self.ctx.find_path("/yolo-nodetypes:conf/leaf-list2"))
        self.assertIsInstance(leaflist2, SLeafList)
        self.assertEqual(leaflist2.min_elements(), 0)
        self.assertEqual(leaflist2.max_elements(), None)

    def test_leaf_list_parsed(self):
        leaflist = next(self.ctx.find_path("/yolo-nodetypes:conf/ratios"))
        self.assertIsInstance(leaflist, SLeafList)
        pnode = leaflist.parsed()
        self.assertIsInstance(pnode, PLeafList)
        self.assertIsNone(next(pnode.musts(), None))
        self.assertIsNone(pnode.when_condition())
        self.assertIsInstance(pnode.type(), PType)
        self.assertIsNone(pnode.units())
        self.assertEqual("2.5", next(pnode.defaults()))
        self.assertEqual(0, pnode.min_elements())
        self.assertIsNone(pnode.max_elements())
        self.assertFalse(pnode.ordered())


# -------------------------------------------------------------------------------------
class BacklinksTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR)
        self.ctx.load_module("yolo-leafref-search")
        self.ctx.load_module("yolo-leafref-search-extmod")

    def tearDown(self):
        self.ctx.destroy()
        self.ctx = None

    def test_backlinks_all_nodes(self):
        expected = [
            "/yolo-leafref-search-extmod:my_extref_list/my_extref",
            "/yolo-leafref-search:refstr",
            "/yolo-leafref-search:refnum",
            "/yolo-leafref-search-extmod:my_extref_list/my_extref_union",
        ]
        refs = self.ctx.find_backlinks_paths()
        expected.sort()
        refs.sort()
        self.assertEqual(expected, refs)

    def test_backlinks_one(self):
        expected = [
            "/yolo-leafref-search-extmod:my_extref_list/my_extref",
            "/yolo-leafref-search:refstr",
            "/yolo-leafref-search-extmod:my_extref_list/my_extref_union",
        ]
        refs = self.ctx.find_backlinks_paths(
            match_path="/yolo-leafref-search:my_list/my_leaf_string"
        )
        expected.sort()
        refs.sort()
        self.assertEqual(expected, refs)

    def test_backlinks_children(self):
        expected = [
            "/yolo-leafref-search-extmod:my_extref_list/my_extref",
            "/yolo-leafref-search:refstr",
            "/yolo-leafref-search:refnum",
            "/yolo-leafref-search-extmod:my_extref_list/my_extref_union",
        ]
        refs = self.ctx.find_backlinks_paths(
            match_path="/yolo-leafref-search:my_list", match_ancestors=True
        )
        expected.sort()
        refs.sort()
        self.assertEqual(expected, refs)

    def test_backlinks_leafref_target_paths(self):
        expected = ["/yolo-leafref-search:my_list/my_leaf_string"]
        refs = self.ctx.find_leafref_path_target_paths(
            "/yolo-leafref-search-extmod:my_extref_list/my_extref"
        )
        expected.sort()
        refs.sort()
        self.assertEqual(expected, refs)


# -------------------------------------------------------------------------------------
class ChoiceTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR)
        self.ctx.load_module("yolo-system")

    def tearDown(self):
        self.ctx.destroy()
        self.ctx = None

    def test_choice_default(self):
        conf = next(self.ctx.find_path("/yolo-system:conf"))
        choice = next(conf.children((SNode.CHOICE,), with_choice=True))
        self.assertIsInstance(choice, SChoice)
        self.assertIsInstance(choice.default(), SCase)

    def test_choice_parsed(self):
        conf = next(self.ctx.find_path("/yolo-system:conf"))
        choice = next(conf.children((SNode.CHOICE,), with_choice=True))
        self.assertIsInstance(choice, SChoice)
        pnode = choice.parsed()
        self.assertIsInstance(pnode, PChoice)

        case_pnode = next(iter(pnode))
        self.assertIsInstance(case_pnode, PCase)
        self.assertIsNotNone(next(iter(case_pnode)))
        self.assertIsNone(case_pnode.when_condition())

        self.assertIsNone(pnode.when_condition())
        self.assertEqual("red", pnode.default())


# -------------------------------------------------------------------------------------
class AnydataTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR)
        self.ctx.load_module("yolo-nodetypes")

    def tearDown(self):
        self.ctx.destroy()
        self.ctx = None

    def test_anydata(self):
        snode = next(self.ctx.find_path("/yolo-nodetypes:any1"))
        self.assertIsInstance(snode, SAnydata)
        assert next(snode.when_conditions()) is not None
        snode2 = next(snode.when_conditions_nodes())
        assert isinstance(snode2, SAnydata)
        assert snode2.cdata == snode.cdata

    def test_anydata_parsed(self):
        snode = next(self.ctx.find_path("/yolo-nodetypes:any1"))
        self.assertIsInstance(snode, SAnydata)
        pnode = snode.parsed()
        self.assertIsInstance(pnode, PAnydata)
        self.assertIsNone(next(pnode.musts(), None))
        self.assertEqual("../cont2", pnode.when_condition())


# -------------------------------------------------------------------------------------
class NotificationTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR)
        self.ctx.load_module("yolo-nodetypes")

    def tearDown(self):
        self.ctx.destroy()
        self.ctx = None

    def test_notification_parsed(self):
        snode = next(self.ctx.find_path("/yolo-nodetypes:cont2"))
        self.assertIsInstance(snode, SContainer)
        pnode = snode.parsed()
        self.assertIsInstance(pnode, PContainer)
        pnode = next(pnode.notifications())
        self.assertIsInstance(pnode, PNotif)
        self.assertIsNone(next(pnode.musts(), None))
        self.assertIsNone(next(pnode.typedefs(), None))
        self.assertIsNone(next(pnode.groupings(), None))
        self.assertIsNotNone(next(iter(pnode)))


# -------------------------------------------------------------------------------------
class IdentityTest(unittest.TestCase):
    def setUp(self):
        self.ctx = Context(YANG_DIR)
        self.module = self.ctx.load_module("yolo-nodetypes")

    def tearDown(self):
        self.ctx.destroy()
        self.ctx = None

    def test_identity_compiled(self):
        sidentity = next(self.module.identities())
        self.assertIsInstance(sidentity, Identity)
        self.assertEqual(sidentity.name(), "base1")
        self.assertEqual(sidentity.description(), "Base 1.")
        self.assertEqual(sidentity.reference(), "Some reference.")
        self.assertIsInstance(sidentity.module(), Module)
        derived = list(sidentity.derived())
        self.assertEqual(2, len(derived))
        for i in derived:
            self.assertIsInstance(i, Identity)
        self.assertEqual(derived[0].name(), "derived1")
        self.assertEqual(derived[1].name(), "derived2")
        self.assertEqual(next(derived[1].extensions()).name(), "identity-name")
        self.assertIsNone(next(sidentity.extensions(), None))
        self.assertIsNone(sidentity.get_extension("ext1"))
        self.assertFalse(sidentity.deprecated())
        self.assertFalse(sidentity.obsolete())
        self.assertEqual("current", sidentity.status())

        snode = next(self.ctx.find_path("/yolo-nodetypes:identity_ref"))
        identities = list(snode.type().identity_bases())
        self.assertEqual(identities[0].name(), sidentity.name())
        self.assertEqual(identities[1].name(), "base2")

    def test_identity_parsed(self):
        pidentity = next(self.module.parsed_identities())
        self.assertIsInstance(pidentity, PIdentity)
        self.assertEqual(pidentity.name(), "base1")
        self.assertIsNone(next(pidentity.if_features(), None))
        self.assertIsNone(next(pidentity.bases(), None))
        self.assertEqual(pidentity.description(), "Base 1.")
        self.assertEqual(pidentity.reference(), "Some reference.")
        self.assertIsNone(next(pidentity.extensions(), None))
        self.assertFalse(pidentity.deprecated())
        self.assertFalse(pidentity.obsolete())
        self.assertEqual("current", pidentity.status())



# Copyright (c) 2018-2019 Robin Jarry
# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: MIT

from contextlib import contextmanager
import logging

from _libyang import ffi, lib
from .util import c2str


# -------------------------------------------------------------------------------------
LOG = logging.getLogger("libyang")
LOG.addHandler(logging.NullHandler())
LOG_LEVELS = {
    lib.LY_LLERR: logging.ERROR,
    lib.LY_LLWRN: logging.WARNING,
    lib.LY_LLVRB: logging.INFO,
    lib.LY_LLDBG: logging.DEBUG,
}


def get_libyang_level(py_level):
    for ly_lvl, py_lvl in LOG_LEVELS.items():
        if py_lvl == py_level:
            return ly_lvl
    return None


@contextmanager
def temp_log_options(opt: int = 0):
    opts = ffi.new("uint32_t *", opt)

    lib.ly_temp_log_options(opts)
    yield
    lib.ly_temp_log_options(ffi.NULL)


@ffi.def_extern(name="lypy_log_cb")
def libyang_c_logging_callback(level, msg, data_path, schema_path, line):
    args = [c2str(msg)]
    fmt = "%s"
    if data_path:
        fmt += ": %s"
        args.append(c2str(data_path))
    if schema_path:
        fmt += ": %s"
        args.append(c2str(schema_path))
    if line != 0:
        fmt += " line %u"
        args.append(line)
    LOG.log(LOG_LEVELS.get(level, logging.NOTSET), fmt, *args)


def configure_logging(enable_py_logger: bool, level: int = logging.ERROR) -> None:
    """
    Configure libyang logging behaviour.

    :arg enable_py_logger:
        If False, configure libyang to store the errors in the context until they are
        consumed when Context.error() is called. This is the default behaviour.

        If True, libyang log messages will be sent to the python "libyang" logger and
        will be processed according to the python logging configuration. Note that by
        default, the "libyang" python logger is created with a NullHandler() which means
        that all messages are lost until another handler is configured for that logger.
    :arg level:
        Python logging level. By default only ERROR messages are stored/logged.
    """
    ly_level = get_libyang_level(level)
    if ly_level is not None:
        lib.ly_log_level(ly_level)
    if enable_py_logger:
        lib.ly_log_options(lib.LY_LOLOG | lib.LY_LOSTORE)
        lib.ly_set_log_clb(lib.lypy_log_cb)
    else:
        lib.ly_log_options(lib.LY_LOSTORE)
        lib.ly_set_log_clb(ffi.NULL)


configure_logging(False, logging.ERROR)
